//+------------------------------------------------------------------+
//|                                                   Optimystic.mqh |
//|                                 Copyright © 2011-2021, Marketeer |
//|                          https://www.mql5.com/en/users/marketeer |
//| MQL's OOP notes: On The Fly Self-Optimization of Expert Advisers |
//|                        https://www.mql5.com/en/blogs/post/683863 |
//| 07/01/2021 All pure virtual methods are replaced with stubs      |
//|            due to latest MQL4 compiler issue with exports        |
//+------------------------------------------------------------------+

enum PERFORMANCE_ESTIMATOR
{
  ESTIMATOR_DEFAULT = -1,     // default
  ESTIMATOR_NET_PROFIT,       // net profit
  ESTIMATOR_PROFIT_FACTOR,    // profit factor
  ESTIMATOR_SHARPE,           // sharpe
  ESTIMATOR_DRAWDOWN_PERCENT, // drawdown (100-DD%)
  ESTIMATOR_WIN_PERCENT,      // win percent
  ESTIMATOR_AVERAGE_PROFIT,   // average profit
  ESTIMATOR_TRADES_COUNT,     // trade count
  ESTIMATOR_CUSTOM            // custom
};

enum TIME_UNIT
{
  UNIT_BAR, // bars
  UNIT_DAY  // days
};

enum OPTIMIZATION_METHOD
{
  METHOD_BRUTE_FORCE,    // brute force
  METHOD_PARTICLE_SWARM  // particle swarm
};

void stub(const string method)
{
  Print(method, ": this is a stub class, don't use it directly, it must be overridden!");
  ExpertRemove();
}

class OptimysticParameter
{
  public:
    virtual string getName() const {stub(__FUNCSIG__); return NULL;}
    virtual double getMin() const {stub(__FUNCSIG__); return NULL;}
    virtual double getMax() const {stub(__FUNCSIG__); return NULL;}
    virtual double getStep() const {stub(__FUNCSIG__); return NULL;}
    virtual double getValue() const {stub(__FUNCSIG__); return NULL;}
    virtual double getBest() const {stub(__FUNCSIG__); return NULL;}
    virtual bool getEnabled() const {stub(__FUNCSIG__); return false;}
    virtual void setEnabled(const bool e) {stub(__FUNCSIG__);}
};

class OptimysticTradeContext
{
  public:
    virtual datetime TimeCurrent() {stub(__FUNCSIG__); return NULL;}
    
    virtual double AccountBalance() {stub(__FUNCSIG__); return NULL;}
    virtual double AccountEquity() {stub(__FUNCSIG__); return NULL;}
    virtual double AccountProfit() {stub(__FUNCSIG__); return NULL;}
    virtual double AccountFreeMargin() {stub(__FUNCSIG__); return NULL;}
    virtual double AccountFreeMarginCheck(string symbol, int cmd, double volume) {stub(__FUNCSIG__); return NULL;}
    virtual double AccountMargin() {stub(__FUNCSIG__); return NULL;}

    virtual int OrderSend(string symbol, int cmd, double volume, double price, int deviation, double stoploss, double takeprofit, string comment = NULL, int magic = 0, datetime expiration = 0, color arrow = clrNONE) {stub(__FUNCSIG__); return NULL;}
    virtual bool OrderClose(int ticket, double lots, double price, int slippage, color arrow = clrNONE) {stub(__FUNCSIG__); return false;}
    virtual bool OrderDelete(int ticket, color arrow = clrNONE) {stub(__FUNCSIG__); return false;}
    virtual bool OrderModify(int ticket, double price, double stoploss, double takeprofit, datetime expiration, color arrow = clrNONE) {stub(__FUNCSIG__); return false;}
    virtual bool OrderSelect(int index, int select, int pool = MODE_TRADES) {stub(__FUNCSIG__); return false;}

    virtual int OrdersHistoryTotal() {stub(__FUNCSIG__); return NULL;}
    virtual int OrdersTotal() {stub(__FUNCSIG__); return NULL;}
    
    virtual double OrderClosePrice() {stub(__FUNCSIG__); return NULL;}
    virtual datetime OrderCloseTime() {stub(__FUNCSIG__); return NULL;}
    virtual string OrderComment() {stub(__FUNCSIG__); return NULL;}
    virtual double OrderCommission() {stub(__FUNCSIG__); return NULL;}
    virtual datetime OrderExpiration() {stub(__FUNCSIG__); return NULL;}
    virtual double OrderLots() {stub(__FUNCSIG__); return NULL;}
    virtual int OrderMagicNumber() {stub(__FUNCSIG__); return NULL;}
    virtual double OrderOpenPrice() {stub(__FUNCSIG__); return NULL;}
    virtual datetime OrderOpenTime() {stub(__FUNCSIG__); return NULL;}
    virtual void OrderPrint() {stub(__FUNCSIG__);}
    virtual double OrderProfit() {stub(__FUNCSIG__); return NULL;}
    virtual double OrderStopLoss() {stub(__FUNCSIG__); return NULL;}
    virtual double OrderSwap() {stub(__FUNCSIG__); return NULL;}
    virtual string OrderSymbol() {stub(__FUNCSIG__); return NULL;}
    virtual double OrderTakeProfit() {stub(__FUNCSIG__); return NULL;}
    virtual int OrderTicket() {stub(__FUNCSIG__); return NULL;}
    virtual int OrderType() {stub(__FUNCSIG__); return NULL;}
    
    virtual double iOpen(string symbol, int tf, int bar) {stub(__FUNCSIG__); return NULL;}
    virtual double iHigh(string symbol, int tf, int bar) {stub(__FUNCSIG__); return NULL;}
    virtual double iLow(string symbol, int tf, int bar) {stub(__FUNCSIG__); return NULL;}
    virtual double iClose(string symbol, int tf, int bar) {stub(__FUNCSIG__); return NULL;}
    virtual datetime iTime(string symbol, int tf, int bar) {stub(__FUNCSIG__); return NULL;}
    virtual int iBarShift(string symbol, int tf, datetime dt, bool exact = false) {stub(__FUNCSIG__); return NULL;}

    // TODO: iVolume iHighest iLowest iBars
    // TODO: iCustom and all standard indicators
    // TODO: MarketInfo
};

class OptimysticCallback
{
  private:
    OptimysticTradeContext *context;
    
  protected:
    datetime TimeCurrent()
    {
      return context != NULL ? context.TimeCurrent() : ::TimeCurrent();
    }
    
    double AccountBalance()
    {
      return context != NULL ? context.AccountBalance() : ::AccountBalance();
    }
    
    double AccountEquity()
    {
      return context != NULL ? context.AccountEquity() : ::AccountEquity();
    }
    
    double AccountProfit()
    {
      return context != NULL ? context.AccountProfit() : ::AccountProfit();
    }
    
    double AccountFreeMargin()
    {
      return context != NULL ? context.AccountFreeMargin() : ::AccountFreeMargin();
    }
    
    double AccountFreeMarginCheck(string symbol, int cmd, double volume)
    {
      return context != NULL ? context.AccountFreeMarginCheck(symbol, cmd, volume) : ::AccountFreeMarginCheck(symbol, cmd, volume);
    }
    
    double AccountMargin()
    {
      return context != NULL ? context.AccountMargin() : ::AccountMargin();
    }
    
    
    int OrdersHistoryTotal()
    {
      return context != NULL ? context.OrdersHistoryTotal() : ::OrdersHistoryTotal();
    }
    
    int OrdersTotal()
    {
      return context != NULL ? context.OrdersTotal() : ::OrdersTotal();
    }
    
    int OrderSend(string symbol, int cmd, double volume, double price, int deviation, double stoploss, double takeprofit, string comment = NULL, int magic = 0, datetime expiration = 0, color arrow = clrNONE)
    {
      return context != NULL ? context.OrderSend(symbol, cmd, volume, price, deviation, stoploss, takeprofit, comment, magic, expiration, arrow) : ::OrderSend(symbol, cmd, volume, price, deviation, stoploss, takeprofit, comment, magic, expiration, arrow);
    }
    
    bool OrderClose(int ticket, double lots, double price, int slippage, color arrow = clrNONE)
    {
      return context != NULL ? context.OrderClose(ticket, lots, price, slippage, arrow) : ::OrderClose(ticket, lots, price, slippage, arrow);
    }
    
    bool OrderDelete(int ticket, color arrow = clrNONE)
    {
      return context != NULL ? context.OrderDelete(ticket, arrow) : ::OrderDelete(ticket, arrow);
    }
    
    bool OrderModify(int ticket, double price, double stoploss, double takeprofit, datetime expiration, color arrow = clrNONE)
    {
      return context != NULL ? context.OrderModify(ticket, price, stoploss, takeprofit, expiration, arrow) : ::OrderModify(ticket, price, stoploss, takeprofit, expiration, arrow);
    }
    
    bool OrderSelect(int index, int select, int pool = MODE_TRADES)
    {
      return context != NULL ? context.OrderSelect(index, select, pool) : ::OrderSelect(index, select, pool);
    }

    double OrderClosePrice()
    {
      return context != NULL ? context.OrderClosePrice() : ::OrderClosePrice();
    }
    
    datetime OrderCloseTime()
    {
      return context != NULL ? context.OrderCloseTime() : ::OrderCloseTime();
    }

    string OrderComment()
    {
      return context != NULL ? context.OrderComment() : ::OrderComment();
    }
    
    double OrderCommission()
    {
      return context != NULL ? context.OrderCommission() : ::OrderCommission();
    }
    
    datetime OrderExpiration()
    {
      return context != NULL ? context.OrderExpiration() : ::OrderExpiration();
    }
    
    double OrderLots()
    {
      return context != NULL ? context.OrderLots() : ::OrderLots();
    }
    
    int OrderMagicNumber()
    {
      return context != NULL ? context.OrderMagicNumber() : ::OrderMagicNumber();
    }
    
    double OrderOpenPrice()
    {
      return context != NULL ? context.OrderOpenPrice() : ::OrderOpenPrice();
    }
    
    datetime OrderOpenTime()
    {
      return context != NULL ? context.OrderOpenTime() : ::OrderOpenTime();
    }
    
    void OrderPrint()
    {
      if(context != NULL) context.OrderPrint();
      else ::OrderPrint();
    }
    
    double OrderProfit()
    {
      return context != NULL ? context.OrderProfit() : ::OrderProfit();
    }
    
    double OrderStopLoss()
    {
      return context != NULL ? context.OrderStopLoss() : ::OrderStopLoss();
    }
    
    double OrderSwap()
    {
      return context != NULL ? context.OrderSwap() : ::OrderSwap();
    }
    
    string OrderSymbol()
    {
      return context != NULL ? context.OrderSymbol() : ::OrderSymbol();
    }
    
    double OrderTakeProfit()
    {
      return context != NULL ? context.OrderTakeProfit() : ::OrderTakeProfit();
    }
    
    int OrderTicket()
    {
      return context != NULL ? context.OrderTicket() : ::OrderTicket();
    }
    
    int OrderType()
    {
      return context != NULL ? context.OrderType() : ::OrderType();
    }

    double iOpen(string symbol, int tf, int bar)
    {
      return context != NULL ? context.iOpen(symbol, tf, bar) : ::iOpen(symbol, tf, bar);
    }

    double iHigh(string symbol, int tf, int bar)
    {
      return context != NULL ? context.iHigh(symbol, tf, bar) : ::iHigh(symbol, tf, bar);
    }

    double iLow(string symbol, int tf, int bar)
    {
      return context != NULL ? context.iLow(symbol, tf, bar) : ::iLow(symbol, tf, bar);
    }

    double iClose(string symbol, int tf, int bar)
    {
      return context != NULL ? context.iClose(symbol, tf, bar) : ::iClose(symbol, tf, bar);
    }
    
    datetime iTime(string symbol, int tf, int bar)
    {
      return context != NULL ? context.iTime(symbol, tf, bar) : ::iTime(symbol, tf, bar);
    }
    
    int iBarShift(string symbol, int tf, datetime dt, bool exact = false)
    {
      return context != NULL ? context.iBarShift(symbol, tf, dt, exact) : ::iBarShift(symbol, tf, dt, exact);
    }
    

  public:
    void setContext(OptimysticTradeContext *c)
    {
      context = c;
    }
    
    OptimysticTradeContext *getContext() const
    {
      return context;
    }
    
    virtual bool onBeforeOptimization() {stub(__FUNCSIG__); return false;}
    virtual bool onApplyParameters() {stub(__FUNCSIG__); return false;}
    virtual void trade(const int bar, const double _Ask, const double _Bid) {stub(__FUNCSIG__);}
    virtual void onReadyParameters()
    {
    }
    virtual double onCustomEstimate()
    {
      return EMPTY_VALUE;
    }
    virtual void onAfterOptimization() {stub(__FUNCSIG__);}
};

class Optimystic
{
  public:
    virtual OptimysticParameter *addParameter(const string name, const double min, const double max, const double step) {stub(__FUNCSIG__); return NULL;}
    virtual OptimysticParameter *getParameter(const int i) const {stub(__FUNCSIG__); return NULL;}
    OptimysticParameter *operator[](int i) const
    {
      return getParameter(i);
    }
    virtual void setHistoryDepth(const int size, const TIME_UNIT unit) {stub(__FUNCSIG__);}
    virtual void setReoptimizationPeriod(const int size, const TIME_UNIT unit) {stub(__FUNCSIG__);}
    virtual void setOptimizationCriterion(const PERFORMANCE_ESTIMATOR type) {stub(__FUNCSIG__);}
    virtual void setDeposit(const double deposit) {stub(__FUNCSIG__);}
    virtual void setSpread(const int points = 0) {stub(__FUNCSIG__);}
    virtual void setCommission(const double c) {stub(__FUNCSIG__);}
    virtual void setEnabled(const bool enabled) {stub(__FUNCSIG__);}
    virtual void setMethod(OPTIMIZATION_METHOD method) {stub(__FUNCSIG__);}
  
    virtual double getOptimalPerformance(PERFORMANCE_ESTIMATOR type = ESTIMATOR_DEFAULT) {stub(__FUNCSIG__); return 0;}
    
    virtual void onTick() {stub(__FUNCSIG__);}
};

#ifndef OPTIMYSTIC_LIBRARY
#import "Optimystic.ex4"
  Optimystic *createInstance(OptimysticCallback &callback);
#import
#endif
